import pytest
from threedi_schema import models

from threedi_modelchecker.checks.timeseries import (
    FirstTimeSeriesEqualTimestepsCheck,
    TimeSeriesEqualTimestepsCheck,
    TimeseriesExistenceCheck,
    TimeseriesIncreasingCheck,
    TimeseriesRowCheck,
    TimeseriesStartsAtZeroCheck,
    TimeseriesTimestepCheck,
    TimeseriesValueCheck,
)

from .factories import BoundaryConditions1DFactory, BoundaryConditions2DFactory


@pytest.mark.parametrize("check_type", ["1d", "2d"])
@pytest.mark.parametrize(
    "timeseries_tuple,expected_invalid",
    [
        ((), 0),  # no timeseries
        (("0,-0.5 \n59,-0.2", "0,-0.5 \n59,-0.2"), 0),  # same levels
        (
            ("0,-0.5 \n59,-0.2", "0,-0.5 \n59,-0.3"),
            0,
        ),  # same timesteps, different levels
        (
            ("0,-0.5 \n59,-0.2", "0,-0.5 \n58,-0.3", "0,-0.5 \n59,-0.3"),
            1,
        ),  # differing timestep, one error
        (
            ("0,-0.5 \n58,-0.2", "0,-0.5 \n59,-0.3", "0,-0.5 \n59,-0.3"),
            2,
        ),  # differing first timestep, all other timeseries error
    ],
)
def test_timeseries_same_timesteps(
    session, timeseries_tuple, check_type, expected_invalid
):
    if check_type == "1d":
        for timeseries in timeseries_tuple:
            BoundaryConditions1DFactory(timeseries=timeseries)
        check = TimeSeriesEqualTimestepsCheck(models.BoundaryCondition1D.timeseries)
    elif check_type == "2d":
        for timeseries in timeseries_tuple:
            BoundaryConditions2DFactory(timeseries=timeseries)
        check = TimeSeriesEqualTimestepsCheck(models.BoundaryConditions2D.timeseries)
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_invalid


@pytest.mark.parametrize(
    "one_d_timeseries_tuple,two_d_timeseries_tuple,expected_invalid",
    [
        ((), (), 0),  # no timeseries
        (("0,-0.5 \n59,-0.2", "0,-0.5 \n59,-0.2"), (), 0),  # no 2d timeseries
        ((), ("0,-0.5 \n59,-0.2", "0,-0.5 \n59,-0.2"), 0),  # no 1d timeseries
        (
            ("0,-0.5 \n59,-0.2", "0,-0.5 \n59,-0.2"),
            ("0,-0.5 \n59,-0.2", "0,-0.5 \n58,-0.2"),
            0,
        ),  # differing second element
        (
            ("0,-0.5 \n59,-0.2", "0,-0.5 \n59,-0.2"),
            ("0,-0.5 \n58,-0.2", "0,-0.5 \n59,-0.2"),
            1,
        ),  # differing first element
    ],
)
def test_first_timeseries_same_timesteps(
    session, one_d_timeseries_tuple, two_d_timeseries_tuple, expected_invalid
):
    for timeseries in one_d_timeseries_tuple:
        BoundaryConditions1DFactory(timeseries=timeseries)
    for timeseries in two_d_timeseries_tuple:
        BoundaryConditions2DFactory(timeseries=timeseries)
    check = FirstTimeSeriesEqualTimestepsCheck()
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_invalid


@pytest.mark.parametrize("timeseries", ["0,-0.5", "0,-0.5 \n59,-0.5\n60,-0.5\n   "])
def test_timeseries_existence_ok(session, timeseries):
    BoundaryConditions2DFactory(timeseries=timeseries)

    check = TimeseriesExistenceCheck(models.BoundaryConditions2D.timeseries)
    invalid = check.get_invalid(session)
    assert len(invalid) == 0


@pytest.mark.parametrize("timeseries", ["", None])
def test_timeseries_existence_error(session, timeseries):
    BoundaryConditions2DFactory(timeseries=timeseries)

    check = TimeseriesExistenceCheck(models.BoundaryConditions2D.timeseries)
    invalid = check.get_invalid(session)
    assert len(invalid) == 1


@pytest.mark.parametrize(
    "timeseries", ["0,-0.5", "0,-0.5 \n59,-0.5\n60,-0.5\n   ", "", None]
)
def test_timeseries_row_check_ok(session, timeseries):
    BoundaryConditions2DFactory(timeseries=timeseries)

    check = TimeseriesRowCheck(models.BoundaryConditions2D.timeseries)
    invalid = check.get_invalid(session)
    assert len(invalid) == 0


@pytest.mark.parametrize("timeseries", ["0,-0.5,14"])
def test_timeseries_row_check_error(session, timeseries):
    BoundaryConditions2DFactory(timeseries=timeseries)

    check = TimeseriesRowCheck(models.BoundaryConditions2D.timeseries)
    invalid = check.get_invalid(session)
    assert len(invalid) == 1


# Note: Invalid rows are 'valid' for this check
@pytest.mark.parametrize(
    "timeseries", ["0,foo", "0,-0.5\n59,-0.5\n60,-0.5", "0,-0.5,14", "", None]
)
def test_timeseries_timestep_check_ok(session, timeseries):
    BoundaryConditions2DFactory(timeseries=timeseries)

    check = TimeseriesTimestepCheck(models.BoundaryConditions2D.timeseries)
    invalid = check.get_invalid(session)
    assert len(invalid) == 0


@pytest.mark.parametrize("timeseries", ["foo,9.1", "-1, 1.0"])
def test_timeseries_timestep_check_error(session, timeseries):
    BoundaryConditions2DFactory(timeseries=timeseries)

    check = TimeseriesTimestepCheck(models.BoundaryConditions2D.timeseries)
    invalid = check.get_invalid(session)
    assert len(invalid) == 1


# Note: Invalid rows are 'valid' for this check
@pytest.mark.parametrize(
    "timeseries",
    [
        "foo,2.1",
        "foo,1E5",
        "foo,-2",
        "0,-0.5 \n59,-0.5\n 60,-0.5\n   ",
        "0,-0.5,14",
        "0,-0.5,14",
        "",
        None,
    ],
)
def test_timeseries_value_check_ok(session, timeseries):
    BoundaryConditions2DFactory(timeseries=timeseries)

    check = TimeseriesValueCheck(models.BoundaryConditions2D.timeseries)
    invalid = check.get_invalid(session)
    assert len(invalid) == 0


@pytest.mark.parametrize("timeseries", ["1,foo", "1,nan", "1,inf", "1,''"])
def test_timeseries_value_check_error(session, timeseries):
    BoundaryConditions2DFactory(timeseries=timeseries)

    check = TimeseriesValueCheck(models.BoundaryConditions2D.timeseries)
    invalid = check.get_invalid(session)
    assert len(invalid) == 1


@pytest.mark.parametrize(
    "timeseries",
    [
        "0,2.1",
        "0,-0.5 \n59, -0.5\n60 ,-0.5\n   ",
        "0,-0.5,14",
        "0,-0.5,14",
        "foo,1.2",
        "1,foo",
        "",
        None,
    ],
)
def test_timeseries_increasing_check_ok(session, timeseries):
    BoundaryConditions2DFactory(timeseries=timeseries)

    check = TimeseriesIncreasingCheck(models.BoundaryConditions2D.timeseries)
    invalid = check.get_invalid(session)
    assert len(invalid) == 0


@pytest.mark.parametrize("timeseries", ["2,1.0\n2,1.0", "2,1.0\n1,1.0\n2,1.0"])
def test_timeseries_increasing_check_error(session, timeseries):
    BoundaryConditions2DFactory(timeseries=timeseries)

    check = TimeseriesIncreasingCheck(models.BoundaryConditions2D.timeseries)
    invalid = check.get_invalid(session)
    assert len(invalid) == 1


@pytest.mark.parametrize(
    "timeseries",
    [
        "0,2.1",
        "0,2.1\n1,4.2",
        "0,2.1\n-1,4.2",
        "0,-0.5 \n59, -0.5\n60 ,-0.5\n   ",
        "0,-0.5,14",
        "0,-0.5,14",
        "foo,1.2",
        "1,foo",
        "",
        None,
    ],
)
def test_timeseries_starts_zero_check_ok(session, timeseries):
    BoundaryConditions2DFactory(timeseries=timeseries)

    check = TimeseriesStartsAtZeroCheck(models.BoundaryConditions2D.timeseries)
    invalid = check.get_invalid(session)
    assert len(invalid) == 0


@pytest.mark.parametrize("timeseries", ["2,1.0", "2,1.0\n3,1.0"])
def test_timeseries_starts_zero_check_err(session, timeseries):
    BoundaryConditions2DFactory(timeseries=timeseries)

    check = TimeseriesStartsAtZeroCheck(models.BoundaryConditions2D.timeseries)
    invalid = check.get_invalid(session)
    assert len(invalid) == 1



from typing import Dict, Iterator, NamedTuple, Optional, Tuple

from threedi_schema import ThreediDatabase

from .checks.base import BaseCheck, CheckLevel
from .checks.raster import LocalContext, ServerContext
from .config import Config

__all__ = ["ThreediModelChecker"]


class ThreediModelChecker:
    def __init__(
        self,
        threedi_db: ThreediDatabase,
        context: Optional[Dict] = None,
        allow_beta_features=False,
    ):
        """Initialize the model checker.

        Optionally, supply the context of the model check:

        - "context_type": "local" or "server", default "local"
        - "raster_interface": a threedi_modelchecker.interfaces.RasterInterface subclass
        - "base_path": (only local) path where to look for rasters (defaults to the db's directory)
        - "available_rasters": (only server) a dict of raster_option -> raster url
        """
        self.db = threedi_db
        self.schema = self.db.schema
        self.schema.validate_schema()
        self.config = Config(
            models=self.models, allow_beta_features=allow_beta_features
        )
        context = {} if context is None else context.copy()
        context_type = context.pop("context_type", "local")
        if context_type == "local":
            context.setdefault("base_path", self.db.base_path)
            self.context = LocalContext(**context)
        elif context_type == "server":
            self.context = ServerContext(**context)
        else:
            raise ValueError(f"Unknown context_type '{context_type}'")

    @property
    def models(self):
        """Returns a list of declared models"""
        return self.schema.declared_models

    def errors(
        self, level=CheckLevel.ERROR, ignore_checks=None
    ) -> Iterator[Tuple[BaseCheck, NamedTuple]]:
        """Iterates and applies checks, returning any failing rows.

        By default, checks of WARNING and INFO level are ignored.

        :return: Tuple of the applied check and the failing row.
        """
        session = self.db.get_session()
        session.model_checker_context = self.context
        for check in self.checks(level=level, ignore_checks=ignore_checks):
            model_errors = check.get_invalid(session)
            for error_row in model_errors:
                yield check, error_row

    def checks(self, level=CheckLevel.ERROR, ignore_checks=None) -> Iterator[BaseCheck]:
        """Iterates over all configured checks

        :return: implementations of BaseChecks
        """
        for check in self.config.iter_checks(level=level, ignore_checks=ignore_checks):
            yield check

    def check_table(self, table):
        pass

    def check_column(self, column):
        pass



from threedi_schema import models

from .base import BaseCheck


def parse_timeseries(timeseries_str):
    if not timeseries_str:
        return []
    output = []
    for line in timeseries_str.split():
        timestep, value = line.split(",")
        timestep = int(timestep.strip())
        output.append([timestep, float(value.strip())])
    return output


def compare_timesteps(first_timeseries: str, second_timeseries: str) -> bool:
    first_timesteps = [pair[0] for pair in parse_timeseries(first_timeseries)]
    second_timesteps = [pair[0] for pair in parse_timeseries(second_timeseries)]
    return first_timesteps == second_timesteps


class TimeseriesExistenceCheck(BaseCheck):
    """Check that an empty timeseries has not been provided."""

    def get_invalid(self, session):
        invalid_rows = []
        for row in self.to_check(session).all():
            # this will catch False, None, "", and any other falsy value
            if not row.timeseries:
                invalid_rows.append(row)

        return invalid_rows

    def description(self):
        return f"{self.column_name} contains an empty timeseries; remove the {self.table.name} instance or provide valid timeseries."


class TimeSeriesEqualTimestepsCheck(BaseCheck):
    """
    Check that the timesteps in all timeseries in a column are equal.

    This checks the timesteps for all timeseries in a column against the timesteps in the first
    timeseries in that column. Consequently, if the first timeseries is wrong, all the other timeseries
    in that column will raise a warning.
    This check does not compare timeseries between different columns; for that, FirstTimeSeriesEqualTimestepsCheck
    is used.
    """

    def get_invalid(self, session):
        invalid_timeseries = []

        first_timeseries = None

        for row in self.to_check(session).all():
            timeseries = row.timeseries

            if not timeseries:
                continue

            if not first_timeseries:
                first_timeseries = timeseries
                continue  # don't compare first timeseries with itself

            try:
                if not compare_timesteps(
                    first_timeseries=first_timeseries,
                    second_timeseries=timeseries,
                ):
                    invalid_timeseries.append(row)
            except Exception:  # other checks will catch these
                pass

        return invalid_timeseries

    def description(self):
        return (
            f"One or more timesteps in {self.column_name} did not match the timesteps in the first timeseries in the column."
            + "The timesteps in all timeseries must be the same."
        )


class FirstTimeSeriesEqualTimestepsCheck(BaseCheck):
    """
    Check that the timesteps in the first timeseries in the boundary condition columns are equal, if they both exist.

    This is used in conjunction with TimeSeriesEqualTimestepsCheck to confirm that the timeseries in all the boundary
    conditions have the same timesteps. If each timeseries within each column has the same timesteps, and the first
    timeseries of each column matches timesteps with the first timeseries of each other column, then the timesteps of
    all the timeseries must be the same.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(column=models.BoundaryCondition1D.timeseries, *args, **kwargs)

    def get_invalid(self, session):
        invalid_timeseries = []

        first_1d_timeseries = (
            session.query(models.BoundaryCondition1D.timeseries.table)
            .order_by(models.BoundaryCondition1D.id)
            .limit(1)
            .all()
        )

        first_2d_timeseries = (
            session.query(models.BoundaryConditions2D.timeseries.table)
            .order_by(models.BoundaryConditions2D.id)
            .limit(1)
            .all()
        )

        if first_1d_timeseries and first_2d_timeseries:
            try:
                if not compare_timesteps(
                    first_timeseries=first_1d_timeseries[0].timeseries,
                    second_timeseries=first_2d_timeseries[0].timeseries,
                ):
                    invalid_timeseries.append(first_1d_timeseries[0])
            except Exception:  # other checks will catch these
                pass

        return invalid_timeseries

    def description(self):
        return (
            "The timesteps for the first v2_1d_boundary_conditions.timeseries did not match the timesteps for the first v2_2d_boundary_conditions.timeseries. "
            + "All boundary conditions must have the same timesteps in their timeseries."
        )


class TimeseriesRowCheck(BaseCheck):
    """Check that each record in a timeserie contains 2 elements"""

    def get_invalid(self, session):
        invalid_timeseries = []

        for row in self.to_check(session).all():
            timeserie = row.timeseries

            if not timeserie:
                continue

            if any(len(x.split(",")) != 2 for x in timeserie.split()):
                invalid_timeseries.append(row)

        return invalid_timeseries

    def description(self):
        return (
            f"{self.column_name} must contain 2 elements per line separated by a comma"
        )


class TimeseriesTimestepCheck(BaseCheck):
    """Check that each record in a timeserie starts with an integer >= 0"""

    def get_invalid(self, session):
        invalid_timeseries = []

        for row in self.to_check(session).all():
            timeserie = row.timeseries

            if not timeserie:
                continue

            for timeseries_row in timeserie.split():
                elems = timeseries_row.split(",")
                if len(elems) != 2:
                    continue  # checked elsewhere

                try:
                    timestep = int(elems[0].strip())
                except ValueError:
                    invalid_timeseries.append(row)
                    continue

                if timestep < 0:
                    invalid_timeseries.append(row)

        return invalid_timeseries

    def description(self):
        return (
            f"{self.column_name} contains an invalid timestep, expected an integer >= 0"
        )


class TimeseriesValueCheck(BaseCheck):
    """Check that each record in a timeserie ends with a float and is not an invalid or empty string"""

    def get_invalid(self, session):
        invalid_timeseries = []

        for row in self.to_check(session).all():
            timeserie = row.timeseries

            if not timeserie:
                continue

            for timeseries_row in timeserie.split():
                elems = timeseries_row.split(",")
                if len(elems) != 2:
                    continue  # checked elsewhere

                try:
                    value = float(elems[1].strip())
                except ValueError:
                    invalid_timeseries.append(row)
                    continue

                if str(value) in {"nan", "inf", "-inf"}:
                    invalid_timeseries.append(row)

        return invalid_timeseries

    def description(self):
        return f"{self.column_name} contains an invalid value, expected a float"


class TimeseriesIncreasingCheck(BaseCheck):
    """The timesteps in a timeseries should increase"""

    def get_invalid(self, session):
        invalid_timeseries = []

        for row in self.to_check(session).all():
            timeserie = row.timeseries
            try:
                timesteps = [x[0] for x in parse_timeseries(timeserie)]
            except (ValueError, TypeError):
                continue  # other checks will catch these

            if len(timesteps) < 2:
                continue

            if not all(b > a for (a, b) in zip(timesteps[:-1], timesteps[1:])):
                invalid_timeseries.append(row)

        return invalid_timeseries

    def description(self):
        return f"{self.column_name} should be monotonically increasing"


class TimeseriesStartsAtZeroCheck(BaseCheck):
    """The timesteps in a timeseries should start at 0"""

    def get_invalid(self, session):
        invalid_timeseries = []

        for row in self.to_check(session).all():
            timeserie = row.timeseries
            try:
                timesteps = [x[0] for x in parse_timeseries(timeserie)]
            except (ValueError, TypeError):
                continue  # other checks will catch these

            if len(timesteps) == 0:
                continue

            if timesteps[0] != 0:
                invalid_timeseries.append(row)

        return invalid_timeseries

    def description(self):
        return f"{self.column_name} should be start at timestamp 0"
